var request = require('request');
var path = require('path');
var os = require('os');
var fs = require('fs');
var exec = require('child_process').exec;
var spawn = require('child_process').spawn;
var ncp = require('ncp');
var del = require('del');
var semver = require('semver');

var unzipLib = require('unzip');

var util = require("util"), EventEmitter = require("events").EventEmitter;
//添加事件监听机制
util.inherits(updater, EventEmitter);
var platform = process.platform;
platform = /^win/.test(platform) ? 'win' : /^darwin/.test(platform) ? 'mac' : 'linux' + (process.arch == 'ia32' ? '32' : '64');
if (platform === "win" && process.arch.indexOf("64") >= 0) {
    platform = "win64";
}

/**
 * Creates new instance of updater. Manifest could be a `package.json` of project.
 *
 * Note that compressed apps are assumed to be downloaded in the format produced by [node-webkit-builder](https://github.com/mllrsohn/node-webkit-builder) (or [grunt-node-webkit-builder](https://github.com/mllrsohn/grunt-node-webkit-builder)).
 *
 * @constructor
 * @param {object} manifest - See the [manifest schema](#manifest-schema) below.
 */
function updater(manifest) {
    this.manifest = manifest;

}


/**
 * Will check the latest available version of the application by requesting the manifest specified in `manifestUrl`.
 *
 * The callback will always be called; the second parameter indicates whether or not there's a newer version.
 * This function assumes you use [Semantic Versioning](http://semver.org) and enforces it; if your local version is `0.2.0` and the remote one is `0.1.23456` then the callback will be called with `false` as the second paramter. If on the off chance you don't use semantic versioning, you could manually download the remote manifest and call `download` if you're happy that the remote version is newer.
 *
 * @param {function} cb - Callback arguments: error, newerVersionExists (`Boolean`), remoteManifest
 */
updater.prototype.checkNewVersion = function (cb) {
    var agent = null;
    if (global.agentOption) {
        agent = global.agentOption;
        request.get(this.manifest.manifestUrl, {proxy: agent}, gotManifest.bind(this)); //get manifest from url
    } else {
        request.get(this.manifest.manifestUrl, gotManifest.bind(this)); //get manifest from url
    }

    /**
     * @private
     */
    function gotManifest(err, req, data) {
        if (err) {
            return cb(err);
        }

        if (req.statusCode < 200 || req.statusCode > 299) {
            return cb(new Error(req.statusCode));
        }

        try {
            data = JSON.parse(data);
        } catch (e) {
            return cb(e)
        }

        version_gt(data, this.manifest, cb);
    }
};

/**
 *
 * @param updateData 从服务器端得到的升级包信息
 * @param pkg  自己的package.json
 * @param callback 回调函数
 */
function version_gt(updateData, pkg, cb) {
    /**
     * 将newVersion分为3位版本号和1位补丁号,补丁号可能为空
     * newVersion只能是3位或者4位
     * @param version
     * @returns {*[]}
     */
    function splitNewVersion(version) {
        var versionArray = version.split(".");
        if ((versionArray.length) == 3) {
            return [version, ""];
        } else if ((versionArray.length) == 4) {
            var index = version.lastIndexOf(".");
            return [version.substring(0, index), version.substring(index + 1)];
        } else {
            global.logger.error("invalid new version:", version);
            throw new Error("invalid new  version:" + version);
        }
    }

    try {
        //如果都有newersion字段，则用新的版本判断方法(4位版本号)
        //判断方式分成2部分判断,先比较前3位，如果相同则比较第4位补丁版本
        if (pkg.newVersion && updateData.newVersion) {
            var pkgVersions = splitNewVersion(pkg.newVersion);
            var updateDataVersions = splitNewVersion(updateData.newVersion);

            if (!semver.eq(updateDataVersions[0], pkgVersions[0])) {  //如果前3位不一样，直接返回结果
                return cb(null, semver.gt(updateDataVersions[0], pkgVersions[0]), updateData);
            }
            else {  //否则比较补丁版本
                return cb(null, new Number(updateDataVersions[1]) > new Number(pkgVersions[1]), updateData);
            }
        }
        //否则继续老版本的判断方式
        return cb(null, semver.gt(updateData.version, pkg.version), updateData);
    } catch (err) {  //版本号不符合规范
        return cb(err);
    }
}


/**
 * Downloads the new app to a template folder
 * @param  {Function} cb - called when download completes. Callback arguments: error, downloaded filepath
 * @param  {Object} newManifest - see [manifest schema](#manifest-schema) below
 * @return {Request} Request - stream, the stream contains `manifest` property with new manifest and 'content-length' property with the size of package.
 */
updater.prototype.download = function (cb, newManifest, progress) {
    var self = this;
    var manifest = newManifest || this.manifest;
    var url = manifest.packages[platform].url;
    var etag = manifest.packages[platform].etag;
    console.log('etag', etag);
    var agent = null;
    if (global.agentOption) {
        agent = global.agentOption;
    }
    var pkg = request(url, {proxy: agent}, function (err, response) {
        if (err) {
            cb(err);
        }
        if (response && (response.statusCode < 200 || response.statusCode >= 300)) {
            pkg.abort();
            return cb(new Error(response.statusCode));
        }
    });
    pkg.on('response', function (response) {
        if (response && response.headers) {
            if(response.headers['content-length']){
                pkg['content-length'] = response.headers['content-length'];
            }else{//未返回值给一个固定值
                pkg['content-length'] = 55*1024*1024;
            }
            pkg['content-loaded'] = 0;
        }
    });
    var filename = path.basename(url);
    if (!path.extname(filename)) {
        filename = filename + ".zip";
    }
    var destinationPath = path.join(os.tmpdir(), filename);
    // download the package to template folder
    fs.unlink(path.join(os.tmpdir(), filename), function () {
        pkg.pipe(fs.createWriteStream(destinationPath));
        pkg.resume();
    });
    pkg.on('error', cb);
    pkg.on('data', function (e) {
        pkg['content-loaded'] += e.length;
        var per = (pkg['content-loaded'] / pkg['content-length']) * 100;
        if(pkg['content-loaded']>pkg['content-length']){
            per = 99;
        }
        per = Math.round(per * 100) / 100.0;
        self.emit('progress', per);
    });

    pkg.on('end', appDownloaded);
    pkg.pause();

    function appDownloaded() {
        setTimeout(function () {
            process.nextTick(function () {
                if (pkg.response.statusCode >= 200 && pkg.response.statusCode < 300) {
                    cb(null, destinationPath, etag);
                }
            });
        }, 2000);
    }

    return pkg;
};


/**
 * Returns executed application path
 * @returns {string}
 */
updater.prototype.getAppPath = function () {
    var appPath = {
        mac: path.join(process.cwd(), '../../..'),
        win: path.dirname(process.execPath)
    };
    appPath.linux32 = appPath.win;
    appPath.linux64 = appPath.win;
    appPath.win64 = appPath.win;
    return appPath[platform];
};


/**
 * Returns current application executable
 * @returns {string}
 */
updater.prototype.getAppExec = function () {
    var execFolder = this.getAppPath();
    var exec = {
        mac: '',
        win: path.basename(process.execPath),
        win64: path.basename(process.execPath),
        linux32: path.basename(process.execPath),
        linux64: path.basename(process.execPath)
    };
    return path.join(execFolder, exec[platform]);
};


/**
 * Will unpack the `filename` in temporary folder.
 * For Windows, [unzip](https://www.mkssoftware.com/docs/man1/unzip.1.asp) is used.
 *
 * @param {string} filename
 * @param {function} cb - Callback arguments: error, unpacked directory
 * @param {object} manifest
 */
updater.prototype.unpack = function (filename, cb, manifest) {
    pUnpack[platform].apply(this, arguments);
};

/**
 * @private
 * @param {string} zipPath
 * @return {string}
 */
var getZipDestinationDirectory = function (zipPath) {
        return path.join(os.tmpdir(), path.basename(zipPath, path.extname(zipPath)));
    },
    /**
     * @private
     * @param {object} manifest
     * @return {string}
     */
    getExecPathRelativeToPackage = function (manifest) {
        var execPath = manifest.packages[platform] && manifest.packages[platform].execPath;

        if (execPath) {
            return execPath;
        }
        else {
            var suffix = {
                win: '.exe',
                win64: ".exe",
                mac: '.app'
            };
            return manifest.name + (suffix[platform] || '');
        }
    };


var pUnpack = {
    /**
     * @private
     */
    mac: function (filename, cb, manifest) {
        var args = arguments,
            extension = path.extname(filename),
            destination = path.join(os.tmpdir(), path.basename(filename, extension));

        if (!fs.existsSync(destination)) {
            fs.mkdirSync(destination);
        }
        logger.info("------unzip upgrade filename:", filename);
        logger.info("------unzip upgrade destination:", destination);
        if (extension === ".zip") {
            exec('unzip -xo "' + filename + '" >/dev/null', {cwd: destination}, function (err) {
                if (err) {
                    console.log(err);
                    return cb(err);
                }
                var appPath = path.join(destination, getExecPathRelativeToPackage(manifest));
                logger.info("-----unzip app path:", appPath);
                cb(null, appPath);
            })

        }
        else if (extension === ".dmg") {
            // just in case if something was wrong during previous mount
            exec('hdiutil unmount /Volumes/' + path.basename(filename, '.dmg'), function (err) {
                // create a CDR from the DMG to bypass any steps which require user interaction
                var cdrPath = filename.replace(/.dmg$/, '.cdr');
                exec('hdiutil convert "' + filename + '" -format UDTO -o "' + cdrPath + '"', function (err) {
                    exec('hdiutil attach "' + cdrPath + '" -nobrowse', function (err) {
                        if (err) {
                            if (err.code == 1) {
                                pUnpack.mac.apply(this, args);
                            }
                            return cb(err);
                        }
                        findMountPoint(path.basename(filename, '.dmg'), cb);
                    });
                });
            });

            function findMountPoint(dmg_name, callback) {
                exec('hdiutil info', function (err, stdout) {
                    if (err) return callback(err);
                    var results = stdout.split("\n");
                    var dmgExp = new RegExp(dmg_name + '$');
                    for (var i = 0, l = results.length; i < l; i++) {
                        if (results[i].match(dmgExp)) {
                            var mountPoint = results[i].split("\t").pop();
                            var fileToRun = path.join(mountPoint, dmg_name + ".app");
                            return callback(null, fileToRun);
                        }
                    }
                    callback(Error("Mount point not found"));
                })
            }
        }
    },
    /**
     * @private
     */
    win: function (filename, cb, manifest) {
        var destinationDirectory = getZipDestinationDirectory(filename);
        var unzip = function(){
        	var readStream = fs.createReadStream(filename);
        	readStream.pipe(unzipLib.Extract({path:destinationDirectory}))
        	.on('finish', ()=>{
        		cb(null, path.join(destinationDirectory, getExecPathRelativeToPackage(manifest)));
        	}).on('error', (err)=>{
        		logger.error("------unzip error:", err);
        	});;
        };
//            unzip = function () {
//                // unzip by C. Spieler (docs: https://www.mkssoftware.com/docs/man1/unzip.1.asp, issues: http://www.info-zip.org/)
//                exec('"' + path.resolve(__dirname, 'tools/unzip.exe') + '" -u -o "' +
//                    filename + '" -d "' + destinationDirectory + '" > NUL', function (err) {
//                    if (err) {
//                        return cb(err);
//                    }
//
//                    cb(null, path.join(destinationDirectory, getExecPathRelativeToPackage(manifest)));
//                });
//            };

        fs.exists(destinationDirectory, function (exists) {
            if (exists) {
                del(destinationDirectory, {force: true}).then(()=>{
                	unzip();
                }, (err) => {
                	 cb(err);
                });
            }
            else {
                unzip();
            }
        });

    },
    /**
     * @private
     */
    linux32: function (filename, cb, manifest) {
        //filename fix
        exec('tar -zxvf "' + filename + '" >/dev/null', {cwd: os.tmpdir()}, function (err) {
            console.log(arguments);
            if (err) {
                console.log(err);
                return cb(err);
            }
            cb(null, path.join(os.tmpdir(), getExecPathRelativeToPackage(manifest)));
        })
    }
};
pUnpack.linux64 = pUnpack.linux32;
pUnpack.win64 = pUnpack.win;

/**
 * Runs installer
 * @param {string} appPath
 * @param {array} args - Arguments which will be passed when running the new app
 * @param {object} options - Optional
 * @returns {function}
 */
updater.prototype.runInstaller = function (appPath, args, options) {
    return pRun[platform].apply(this, arguments);
};

var pRun = {
    /**
     * @private
     */
    mac: function (appPath, args, options) {
        //spawn
        if (args && args.length) {
            args = [appPath].concat('--args', args);
        } else {
            args = [appPath];
        }
        return run('open', args, options);
    },
    /**
     * @private
     */
    win: function (appPath, args, options, cb) {
        return run(appPath, args, options, cb);
    },
    /**
     * @private
     */
    linux32: function (appPath, args, options, cb) {
        var appExec = path.join(appPath, path.basename(this.getAppExec()));
        fs.chmodSync(appExec, 0o755)
        if (!options) options = {};
        options.cwd = appPath;
        return run(appPath + "/" + path.basename(this.getAppExec()), args, options, cb);
    }
};

pRun.linux64 = pRun.linux32;
pRun.win64 = pRun.win;

/**
 * @private
 */
function run(path, args, options) {
    var opts = {
        detached: true
    };
    for (var key in options) {
        opts[key] = options[key];
    }
    logger.info("sparn path,args", path, args);
    var sp = spawn(path, args, opts);
    sp.unref();
    return sp;
}

/**
 * Installs the app (copies current application to `copyPath`)
 * @param {string} copyPath
 * @param {function} cb - Callback arguments: error
 */
updater.prototype.install = function (copyPath, cb) {
    pInstall[platform].apply(this, arguments);
};

var pInstall = {
    /**
     * @private
     */
    mac: function (to, cb) {
        ncp(this.getAppPath(), to, cb);
    },
    /**
     * @private
     */
    win: function (to, cb) {
        var self = this;
        appDeleted();

        function appDeleted() {
            ncp(self.getAppPath(), to, {filter: filter}, appCopied);
        }

        function deleteApp(cb) {
            del(to, {force: true}, cb).then(()=>{
            	cb();
            }, (err) => {
            	 cb(err);
            });
        }

        function appCopied(err) {
            if (err) {
                setTimeout(appDeleted, 100);
                return
            }
            cb();
        }

        function filter(source1) {//过滤不用更新的文件夹
            var stats = fs.lstatSync(source1);
            if (path.normalize(source1) == path.normalize(self.getAppPath())) {//根目录直接跳过
                return true;
            }
            if (stats.isDirectory()) {//如果是文件夹，只升级language目录
                if (source1.indexOf(path.normalize(self.getAppPath()) + '/language') > -1 || source1.indexOf(path.normalize(self.getAppPath()) + '/client_ui') > -1 || source1.indexOf(path.normalize(self.getAppPath()) + '/locales') > -1) {
                    return true;
                } else {
                    return false;
                }
            } else {//根目录下的所有文件全部升级
                return true;
            }
        }
    },
    /**
     * @private
     */
    linux32: function (to, cb) {
        ncp(this.getAppPath(), to, cb);
    }
};
pInstall.linux64 = pInstall.linux32;
pInstall.win64 = pInstall.win;

/**
 * Runs the app from original app executable path.
 * @param {string} execPath
 * @param {array} args - Arguments passed to the app being ran.
 * @param {object} options - Optional. See `spawn` from nodejs docs.
 *
 * Note: if this doesn't work, try `gui.Shell.openItem(execPath)` (see [node-webkit Shell](https://github.com/rogerwang/node-webkit/wiki/Shell)).
 */
updater.prototype.run = function (execPath, args, options) {
    var arg = arguments;
    if (platform.indexOf('linux') === 0) arg[0] = path.dirname(arg[0]);
    pRun[platform].apply(this, arg);
};

module.exports = updater;
